@@ -31,14 +31,21 @@ import javax.net.ssl.*;
 import javax.net.ssl.SSLServerSocketFactory;
 import java.security.KeyStore;
 import java.security.PrivateKey;
+import java.security.KeyPair;
+import java.security.KeyPairGenerator;
+import java.security.Signature;
 import java.security.cert.X509Certificate;
 import java.security.SecureRandom;
 import java.math.BigInteger;
 import java.io.FileOutputStream;
 
 // internal sun APIs; warnings galore
-import sun.security.x509.*;
 import java.security.cert.*;
+import sun.security.x509.*;
+
+// for Java 8, CertKeyGen is located in a different package
+// 
+//import sun.security.tools.keytool.CertAndKeyGen;
 
 /** 
  *  Encapsulate Socket, SSL, communication options, parameters, etc.
@@ -54,7 +61,7 @@ public class SocketOpts {
    public static final int defaultConnectTimeout = 3000; /* ms */
    public static int defaultMaxBacklog = 50;
    public static boolean defaultReuseAddr = true;
-   public static int defaultReadTimeout = 15*1000; /* ms */
+   public static final int defaultReadTimeout = 15*1000; /* ms */
 
    private static final int keysize=1024;
    private static final String commonName = "FQDN";
@@ -105,20 +112,34 @@ public class SocketOpts {
 
        // load up our Cartographer keystore for evaluating
        // our certs and create a Trust manager
-       ks = KeyStore.getInstance("jks");
+       if (isAndroid()) {
+          ks = KeyStore.getInstance("BKS");
+       }
+       else {
+          ks = KeyStore.getInstance("jks");
+       }
        //ks1 = KeyStore.getInstance("jks");
        
        if (cl != null) {
-          kis = cl.getResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
-          //kis1 = cl.getResourceAsStream("org/krupczak/xmp/Cartographerkey.keystore");
-          if (kis == null) {
-             kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
-             //kis1 = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographerkey.keystore");
-          }
+          if (isAndroid()) {
+             kis = cl.getResourceAsStream("org/krupczak/xmp/Cartographer.bks");
+             if (kis == null) {
+                kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.bks");
+	     }
+	  } else {
+             kis = cl.getResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
+             if (kis == null) {
+                kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
+	     }
+	  }
        }
        else {
-          kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
-          //kis1 = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographerkey.keystore");
+	   if (isAndroid()) {
+              kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.bks");
+	   }
+	   else {
+              kis = ClassLoader.getSystemResourceAsStream("org/krupczak/xmp/Cartographer.keystore");
+	   }
        }
 
        // if we dont pull it out of our JAR, try it in some well-known
@@ -147,7 +168,13 @@ public class SocketOpts {
        //ks.load(fis,password);
        //fis.close();
 
-       TrustManagerFactory tmf = TrustManagerFactory.getInstance("SunX509");
+       TrustManagerFactory tmf;
+       if (isAndroid()) {
+          tmf = TrustManagerFactory.getInstance("X509");
+       }
+       else {
+          tmf = TrustManagerFactory.getInstance("SunX509");
+       }
        tmf.init(ks);
        tm = tmf.getTrustManagers();
        keystoreFound = true;
@@ -156,7 +183,8 @@ public class SocketOpts {
        km = null;
        //km = generateSignedCredentials(gethostname(),ks,ks1);
 
-       km = generateSignedCredentials(gethostname(),ks,ks);
+       if (isAndroid() == false) 
+          km = generateSignedCredentials(gethostname(),ks,ks);
 
      } catch (Exception e) {
         ks = null;
@@ -165,11 +193,15 @@ public class SocketOpts {
         System.out.println("SocketOpts: Exception "+e.getMessage());
         e.printStackTrace();
 
-        // this sets our Certificate authority 
-        System.setProperty("javax.net.ssl.trustStore","Cartographer.keystore");
+        // this sets our Certificate authority
+	if (isAndroid()) {
+           System.setProperty("javax.net.ssl.trustStore","Cartographer.bks");
+	}
+	else {
+           System.setProperty("javax.net.ssl.trustStore","Cartographer.keystore");
+	}
         System.setProperty("javax.net.ssl.trustStorePassword","cartographer");
         System.setProperty("javax.net.debug","ssl");
-
         keystoreFound = false;
      }
 
@@ -225,9 +257,12 @@ public class SocketOpts {
    private KeyManager[] generateSelfSignedCredentials(String hostname)
    {
        KeyStore keyStore;
-       CertAndKeyGen keypair;
+       //CertAndKeyGen keypair;
+       KeyPair keypair;
+       KeyPairGenerator kpgen;
        Date notBefore, notAfter;
        Calendar cal = Calendar.getInstance();
+       Signature rsa;
 
        try {
          
@@ -236,8 +271,14 @@ public class SocketOpts {
          keyStore.load(null,null);
 
          // generate a key pair
-         keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
-         keypair.generate(keysize);
+         //keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
+         //keypair.generate(keysize);
+         kpgen = KeyPairGenerator.getInstance("RSA");
+         kpgen.initialize(keysize);
+         keypair = kpgen.generateKeyPair();
+
+         rsa = Signature.getInstance("SHA1withRSA");
+         rsa.initSign(keypair.getPrivate());
 
          // generate x500 name
          // we'd like to add extension FQND,hostname XXX
@@ -255,14 +296,18 @@ public class SocketOpts {
          notAfter = cal.getTime();
          X509Certificate[] chain = new X509Certificate[1];
 
-         chain[0] = keypair.getSelfCertificate(x500Name,
-                                             notBefore,
-                                             5*365*24*60*60 //secs?
-                                             );
+         // can't call getSelfCertificate() on keypair
+         // only on CertAndKeyGen which is deprecated in Java 8
+         // there is code to generate self-signed cert
+         // out on net but we don't need it
+         // chain[0] = keypair.getSelfCertificate(x500Name,
+         //                                    notBefore,
+         //                                    5*365*24*60*60 //secs?
+         //                                    );
 
          // stick keys and cert in the keystore
          keyStore.setKeyEntry("xmpentity",
-                            (PrivateKey)keypair.getPrivateKey(),
+                            (PrivateKey)keypair.getPrivate(),
                             password,
                             chain);
 
@@ -311,7 +356,9 @@ public class SocketOpts {
                                                   KeyStore caPrivKeyStore)
    {
        KeyStore keyStore;
-       CertAndKeyGen keypair;
+       //CertAndKeyGen keypair;
+       KeyPair keypair;
+       KeyPairGenerator kpgen;
        Date notBefore, notAfter;
        Calendar cal = Calendar.getInstance();
        String algorithm = new String("SHA1withRSA");
@@ -336,8 +383,12 @@ public class SocketOpts {
          keyStore.load(null,null);
 
          // generate a key pair
-         keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
-         keypair.generate(keysize);
+         //keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
+         //keypair.generate(keysize);
+         kpgen = KeyPairGenerator.getInstance("RSA");
+         kpgen.initialize(keysize);
+         keypair = kpgen.generateKeyPair();
+
 
          // generate the certificate
          X509CertInfo info = new X509CertInfo();
@@ -354,7 +405,7 @@ public class SocketOpts {
          // issuer is our Cartographer CA
          info.set(X509CertInfo.ISSUER,new CertificateIssuerName(issuer));
 
-         info.set(X509CertInfo.KEY,new CertificateX509Key(keypair.getPublicKey()));
+         info.set(X509CertInfo.KEY,new CertificateX509Key(keypair.getPublic()));
          info.set(X509CertInfo.VERSION,new CertificateVersion(CertificateVersion.V3));
          AlgorithmId algo = new AlgorithmId(AlgorithmId.md5WithRSAEncryption_oid);
          info.set(X509CertInfo.ALGORITHM_ID, new CertificateAlgorithmId(algo));
@@ -374,14 +425,14 @@ public class SocketOpts {
          // sign the cert to identify the algorithm thats used
          X509CertImpl cert = new X509CertImpl(info);
          cert.sign(caPrivateKey, algorithm);
-         //cert.sign(bogusKeypair.getPrivateKey(), algorithm);
+         //cert.sign(bogusKeypair.getPrivate(), algorithm);
  
          // Update the algorith, and resign.
          algo = (AlgorithmId)cert.get(X509CertImpl.SIG_ALG);
          info.set(CertificateAlgorithmId.NAME + "." + 
                   CertificateAlgorithmId.ALGORITHM,algo);
          cert = new X509CertImpl(info);
-         //cert.sign(bogusKeypair.getPrivateKey(), algorithm);
+         //cert.sign(bogusKeypair.getPrivate(), algorithm);
          cert.sign(caPrivateKey, algorithm);
 
          X509Certificate[] chain = new X509Certificate[1];
@@ -389,7 +440,7 @@ public class SocketOpts {
 
          // store them in keystore
          keyStore.setKeyEntry("XmpEntity",
-                            (PrivateKey)keypair.getPrivateKey(),
+                            (PrivateKey)keypair.getPrivate(),
                             password,
                             chain);
 
@@ -417,7 +468,9 @@ public class SocketOpts {
                                                   KeyStore caPrivKeyStore)
    {
        KeyStore keyStore;
-       CertAndKeyGen keypair;
+       //CertAndKeyGen keypair;
+       KeyPair keypair;
+       KeyPairGenerator kpgen;
        Date notBefore, notAfter;
        Calendar cal = Calendar.getInstance();
        String algorithm = new String("SHA1withRSA");
@@ -443,8 +496,11 @@ public class SocketOpts {
          keyStore.load(null,null);
 
          // generate a key pair
-         keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
-         keypair.generate(keysize);
+         // keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
+         // keypair.generate(keysize);
+         kpgen = KeyPairGenerator.getInstance("RSA");
+         kpgen.initialize(keysize);
+         keypair = kpgen.generateKeyPair();
 
          // generate the certificate request
          X509CertInfo info = new X509CertInfo();
@@ -455,18 +511,57 @@ public class SocketOpts {
          notBefore = new Date(cal.getTimeInMillis() - 86400000l);
          cal.add(Calendar.YEAR,5);
          notAfter = new Date(notBefore.getTime() + 5*365 * 86400000l);
-         CertificateValidity interval = new CertificateValidity(notBefore,notAfter);
+         CertificateValidity interval = 
+                             new CertificateValidity(notBefore,notAfter);
          BigInteger sn = new BigInteger(64, new SecureRandom());
          X500Name owner = new X500Name(dn); // my Xmp protocol entity
+         CertificateSubjectName csn = new CertificateSubjectName(owner);
+
+         //System.out.println("CertificateSubjectName "+csn);
+         //System.out.println("CertificateSubjectName class "+csn.getClass());
+         //System.out.println("X500Name "+owner);
+         //System.out.println("X500Name is class "+owner.getClass());
+
          info.set(X509CertInfo.VALIDITY,interval);
          info.set(X509CertInfo.SERIAL_NUMBER,new CertificateSerialNumber(sn));
-         info.set(X509CertInfo.SUBJECT,new CertificateSubjectName(owner));
+
+         try {
+           info.set(X509CertInfo.SUBJECT,csn);
+         } catch (Exception xE) {
+	     //System.out.println("Exception setting Certificate Subject");
+             //System.out.println(xE);
+             //System.out.println("Trying again with X500Name");
+             info.set(X509CertInfo.SUBJECT,owner);
+             //System.out.println("Retry with X500Name seemed to work");
+         }
 
          // issuer is our Cartographer CA
-         info.set(X509CertInfo.ISSUER,new CertificateIssuerName(issuer));
+         try {
+	   info.set(X509CertInfo.ISSUER,new CertificateIssuerName(issuer));
+         } catch (Exception xE) {
+	     //System.out.println("Exception setting Certificate Issuer");
+             //System.out.println(xE);
+             //System.out.println("Trying again with X500Name");
+  	     info.set(X509CertInfo.ISSUER,issuer);
+             //System.out.println("Retry with X500Name seemed to work");
+         }
+
+         try {
+           info.set(X509CertInfo.KEY,
+                    new CertificateX509Key(keypair.getPublic()));
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate Key");
+             System.out.println(xE);
+         }
+
+         try {
+           info.set(X509CertInfo.VERSION,
+                    new CertificateVersion(CertificateVersion.V3));
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate version");
+             System.out.println(xE);
+         }
 
-         info.set(X509CertInfo.KEY,new CertificateX509Key(keypair.getPublicKey()));
-         info.set(X509CertInfo.VERSION,new CertificateVersion(CertificateVersion.V3));
          AlgorithmId algo = new AlgorithmId(AlgorithmId.md5WithRSAEncryption_oid);
          info.set(X509CertInfo.ALGORITHM_ID, new CertificateAlgorithmId(algo));
 
@@ -485,7 +580,7 @@ public class SocketOpts {
          // generate and sign cert request to identify the algorithm thats used
          X509CertImpl cert = new X509CertImpl(info);
          cert.sign(caPrivateKey, algorithm);
-         //cert.sign(keypair.getPrivateKey(), algorithm);
+         //cert.sign(keypair.getPrivate, algorithm);
  
          // Update the algorithm, and resign with CA cert private key
          // for the real certificate
@@ -500,7 +595,153 @@ public class SocketOpts {
 
          // store them in keystore
          keyStore.setKeyEntry("XmpEntity",
-                            (PrivateKey)keypair.getPrivateKey(),
+                            (PrivateKey)keypair.getPrivate(),
+                            password,
+                            chain);
+
+         // write new keystore?
+         //writeNewKeystore(caCert,caPrivateKey);
+
+         // make keymanager out of keystore
+         KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
+         kmf.init(keyStore,password);
+
+         KeyManager km[];
+         km = kmf.getKeyManagers();
+
+         return km;
+
+       } catch (Exception e) { 
+           System.out.println("generateSignedCredentials: exception "+e); 
+           e.printStackTrace();
+           return null; 
+       }
+   }
+
+   private KeyManager[] generateSignedCredentialsNew(String fqdn, 
+                                                  KeyStore caKeyStore,
+                                                  KeyStore caPrivKeyStore)
+   {
+       KeyStore keyStore;
+       //CertAndKeyGen keypair;
+       KeyPair keypair;
+       KeyPairGenerator kpgen;
+       Date notBefore, notAfter;
+       Calendar cal = Calendar.getInstance();
+       String algorithm = new String("SHA1withRSA");
+
+       try {
+
+         // get info from our CA keystore
+         Certificate caCert = caKeyStore.getCertificate("ca");
+         byte[] encoded = caCert.getEncoded();
+         X509CertImpl caCertImpl = new X509CertImpl(encoded);
+         X509CertInfo caCertInfo = (X509CertInfo)caCertImpl.get(X509CertImpl.NAME+"."+X509CertImpl.INFO);
+         X500Name issuer = (X500Name)caCertInfo.get(X509CertInfo.SUBJECT+"."+CertificateIssuerName.DN_NAME);
+
+         // get private key from our key store
+	 PrivateKey caPrivateKey = (PrivateKey)caPrivKeyStore.getKey("ca",
+                                                                     password);
+
+         // distinguished name
+         String dn = "C=US, ST=Georgia, L=Smyrna, O=Krupczak.org, OU=Cartographer, CN="+fqdn;
+         
+         // create empty keystore
+         keyStore = KeyStore.getInstance("JKS");
+         keyStore.load(null,null);
+
+         // generate a key pair
+         // keypair = new CertAndKeyGen("RSA","SHA1WithRSA",null);
+         // keypair.generate(keysize);
+         kpgen = KeyPairGenerator.getInstance("RSA");
+         kpgen.initialize(keysize);
+         keypair = kpgen.generateKeyPair();
+
+         // generate the certificate request
+         X509CertInfo info = new X509CertInfo();
+
+         // make notBefore=yesterday to deal with vagaries of 
+         // clock synchronization and GMT offsets and
+         // computers out of sync
+         notBefore = new Date(cal.getTimeInMillis() - 86400000l);
+         cal.add(Calendar.YEAR,5);
+         notAfter = new Date(notBefore.getTime() + 5*365 * 86400000l);
+         CertificateValidity interval = 
+                             new CertificateValidity(notBefore,notAfter);
+         BigInteger sn = new BigInteger(64, new SecureRandom());
+         X500Name owner = new X500Name(dn); // my Xmp protocol entity
+         CertificateSubjectName csn = new CertificateSubjectName(owner);
+
+         System.out.println("CertificateSubjectName "+csn);
+         System.out.println("CertificateSubjectName is class "+csn.getClass());
+         info.set(X509CertInfo.VALIDITY,interval);
+         info.set(X509CertInfo.SERIAL_NUMBER,new CertificateSerialNumber(sn));
+
+         try {
+           info.set(X509CertInfo.SUBJECT,csn);
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate Subject");
+             System.out.println(xE);
+         }
+
+         // issuer is our Cartographer CA
+         try {
+	   info.set(X509CertInfo.ISSUER,new CertificateIssuerName(issuer));
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate Issuer");
+             System.out.println(xE);
+         }
+
+         try {
+           info.set(X509CertInfo.KEY,
+                    new CertificateX509Key(keypair.getPublic()));
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate Key");
+             System.out.println(xE);
+         }
+
+         try {
+           info.set(X509CertInfo.VERSION,
+                    new CertificateVersion(CertificateVersion.V3));
+         } catch (Exception xE) {
+	     System.out.println("Exception setting Certificate version");
+             System.out.println(xE);
+         }
+
+         AlgorithmId algo = new AlgorithmId(AlgorithmId.md5WithRSAEncryption_oid);
+         info.set(X509CertInfo.ALGORITHM_ID, new CertificateAlgorithmId(algo));
+
+         // set Subject Alt Name extension and others if applicable
+         CertificateExtensions ext = new CertificateExtensions();
+         DNSName dnsName = new DNSName(fqdn);
+         GeneralName generalName = new GeneralName(dnsName);
+         GeneralNames generalNames = new GeneralNames();
+         generalNames.add(generalName);
+         ext.set(SubjectAlternativeNameExtension.NAME,
+		 new SubjectAlternativeNameExtension(generalNames));
+         ext.set(BasicConstraintsExtension.NAME, 
+                 new BasicConstraintsExtension(false,1));
+         info.set(X509CertInfo.EXTENSIONS,ext);
+
+         // generate and sign cert request to identify the algorithm thats used
+         X509CertImpl cert = new X509CertImpl(info);
+         cert.sign(caPrivateKey, algorithm);
+         //cert.sign(keypair.getPrivate, algorithm);
+ 
+         // Update the algorithm, and resign with CA cert private key
+         // for the real certificate
+         algo = (AlgorithmId)cert.get(X509CertImpl.SIG_ALG);
+         info.set(CertificateAlgorithmId.NAME + "." + 
+                  CertificateAlgorithmId.ALGORITHM,algo);
+         cert = new X509CertImpl(info);
+         cert.sign(caPrivateKey, algorithm);
+
+         X509Certificate[] chain = new X509Certificate[1];
+         chain[0] = cert;
+
+         // store them in keystore
+         keyStore.setKeyEntry("XmpEntity",
+                            (PrivateKey)keypair.getPrivate(),
                             password,
                             chain);
 
@@ -570,4 +811,33 @@ public class SocketOpts {
    // set the cert that we are using for server sessions
    // this cert is signed by our Cartographer CA
 
+   // are we running on Android?
+   public boolean isAndroid()
+   {
+       String runtime;
+
+       runtime = System.getProperty("java.runtime.name");
+       if (runtime.contains("android") == true)
+	   return true;
+       if (runtime.contains("Android") == true)
+	   return true;
+       if (runtime.contains("ANDROID") == true)
+	   return true;
+
+       return false;
+       
+   } // isAndroid()
+
+   public boolean isSunJava()
+   {
+       String runtime;
+
+       runtime = System.getProperty("java.runtime.name");
+       if (runtime.contains("Java(TM)") == true)
+	   return true;
+
+       return false;
+
+   } // isSunJava()
+
 } /* class SocketOpts */
